==> cego-2.5.13.tar.gz: cego/src/CegoDbThreadPool.cc <==
///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoDbThreadPool.cc  
// -------------------                                                     
// Cego db threadpool class implementation
//                                                         
// Design and Implementation by Bjoern Lemke
//               
// (C)opyright 2000-2010 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoDbThreadPool
// 
// Description: 
//
// Status: QG-2.5.5
//
///////////////////////////////////////////////////////////////////////////////

#include "Exception.h"
#include "Host.h"
#include "NetHandler.h"
#include "Net.h"
#include "Sleeper.h"
#include "ThreadLock.h"

#include "XMLSuite.h"

#include "CegoXMLdef.h"
#include "CegoDbThreadPool.h"

#include "CegoDefs.h"

static ThreadLock queueLock("DBQUEUE"); 

CegoDbThreadPool::CegoDbThreadPool() : Thread()
{
    queueLock.init(LCKMNG_LOCKWAITDELAY, LCKMNG_COUNTSTAT);

}

CegoDbThreadPool::CegoDbThreadPool(int poolLimit, CegoDatabaseManager *pDBMng)  : Thread()
{
    queueLock.init(LCKMNG_LOCKWAITDELAY, LCKMNG_COUNTSTAT);

    _poolLimit = poolLimit;

    pDBMng->getDataPort(_dataPortNo);

    // _dataPortNo = pDoc->getRootElement()->getAttributeValue(XML_DBPORT_ATTR).asInteger();

    _threadId = (long*)malloc(_poolLimit * sizeof(long));
    _numRequest = (long*)malloc(_poolLimit * sizeof(long));
    _threadState = (ThreadState*)malloc(_poolLimit * sizeof(ThreadState));
    _threadList = (CegoDbThread**)malloc(_poolLimit * sizeof(CegoDbThread*));

    _terminated=false;

    long i=0;
    while ( i < _poolLimit )
    {
	
	_threadState[i] = READY;
	_threadList[i]= new CegoDbThread(this, pDBMng);
	_numRequest[i]=0;
	_threadId[i]=i;
	_threadList[i]->start(&_threadId[i]);
	
	i++;
    }
    
    _pDBMng = pDBMng;

    _modId = _pDBMng->getModId("CegoDbThreadPool");
}


CegoDbThreadPool::~CegoDbThreadPool()  
{
    _terminated=true;

    long i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->abortSession();	
	i++;
    }

    _joined=false;
    int count=0;
    while ( _joined == false && count < POOL_TERMWAIT )
    {
	Sleeper s;
	s.secSleep(1);
	count++;
    }
    
    if ( _joined )
    {
	_pDBMng->log(_modId, Logger::NOTICE, Chain("All db threads terminated"));
	join(getTid());
    }
    else
    {
	_pDBMng->log(_modId, Logger::NOTICE, Chain("Canceling hanging db sessions ..."));
	cancel();
    }
    i=0;
    while ( i < _poolLimit )
    {
	delete _threadList[i];
	i++;
    }
    delete _numRequest;
    delete _threadId;
    delete _threadState;
	    
}

void  CegoDbThreadPool::loadObjects(int tabSetId)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->loadObjects(tabSetId);
	i++;
    }
}

void  CegoDbThreadPool::unloadObjects(int tabSetId)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->unloadObjects(tabSetId);
	i++;
    }
}

void CegoDbThreadPool::removeObject(int tabSetId, const Chain& objName, CegoObject::ObjectType type)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->removeObject(tabSetId, objName, type);
	i++;
    }    
}


void CegoDbThreadPool::loadObject(int tabSetId, const Chain& objName, CegoObject::ObjectType type)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->loadObject(tabSetId, objName, type);
	i++;
    }    
}

int CegoDbThreadPool::getPoolLimit() const
{
    return _poolLimit;
}

void CegoDbThreadPool::getThreadInfo(long i, long& numRequest, CegoDbThreadPool::ThreadState& state, Chain& action)
{
    state = _threadState[i];
    numRequest = _numRequest[i];    
    action = _threadList[i]->lastAction();
}

void CegoDbThreadPool::setState(long i, CegoDbThreadPool::ThreadState state)
{
    _threadState[i] = state;
}

void CegoDbThreadPool::incNumRequest(long i)
{
    _numRequest[i]++;
}

void CegoDbThreadPool::setTid(long i, long tid)
{
    _threadId[i] = tid;
}

void  CegoDbThreadPool::abortThread(long i)
{
    _threadList[i]->abortSession();
}

void* CegoDbThreadPool::job(void* arg)
{

    try {
    	
	
	Net net( NETMNG_MSG_BUFLEN );
	
	// cout << "Serving portno " << _portNo << endl;
	net.serve(_dataPortNo);
	
	while ( ! _terminated ) 
	{
	    
	    // take the mutex to avoid cpu load during wait position
	
	    bool noReq=false;
	    lockQueue();
	    if ( _requestQueue.Size() == 0  )  noReq=true;
	    unlockQueue();
	    	    
	    if ( noReq )
		lockQueue();
	    
	    NetHandler* pHandle;

	    try {

		pHandle = net.nextRequest( NETMNG_SELECT_TIMEOUT );

	    }
	    catch ( Exception e )
	    {
		Chain msg;
		e.pop(msg);
		_pDBMng->log(_modId, Logger::LOGERR, Chain("Error on incoming connection : ") + msg);
	    }
	    
	    if ( noReq )
	    {
		unlockQueue();
		
		// struct timespec delay;	
		// delay.tv_sec = 0; 
		// delay.tv_nsec = NETMNG_QUEUE_DELAY;
		// nanosleep(&delay, NULL);

		Sleeper s;
		s.nanoSleep(NETMNG_QUEUE_DELAY);
	    }
	    
	    if ( pHandle )
	    {
		// pHandle->setSendTimeout(NETMNG_SEND_TIMEOUT);
		// pHandle->setRecvTimeout(NETMNG_RECV_TIMEOUT);

		lockQueue();
		_requestQueue.Insert(pHandle);
		unlockQueue();
	    }
	}
	long i=0;
	while ( i < _poolLimit )
	{
	    _pDBMng->log(_modId, Logger::DEBUG, Chain("Waiting for db thread with tid ") + Chain(_threadList[i]->getTid()) + Chain(" to terminate"));
	    join(_threadList[i]->getTid());
	    i++;
	}
	_joined=true;
    }
    catch ( Exception e)
    {
	Chain msg;
	e.pop(msg);	
	_pDBMng->log(_modId, Logger::LOGERR, Chain("Db Thread start failed : ") +  msg);	
	_terminated=true;
    }

    // waiting for db threads to terminate
    
    return 0;
}

bool CegoDbThreadPool::isTerminated()
{
    return _terminated;
}

NetHandler* CegoDbThreadPool::nextRequest()
{

    lockQueue();

    NetHandler **pRequest = _requestQueue.First();

    if ( pRequest )
    {

	NetHandler *pN = *pRequest;
	_requestQueue.Remove(*pRequest);
	unlockQueue();

	return pN; 
    }
    else
    {
	unlockQueue();
	return 0;
    }    
}

void CegoDbThreadPool::lockQueue()
{
    queueLock.writeLock();
}

void CegoDbThreadPool::unlockQueue()
{
    queueLock.unlock();
}


void CegoDbThreadPool::getLockStat(Chain& lockName, long& lockCount, long &numRdLock, long &numWrLock, long &sumRdDelay, long &sumWrDelay)
{
    lockName = queueLock.getId();
    lockCount = queueLock.numLockTry();

    numRdLock = queueLock.numReadLock();
    numWrLock = queueLock.numWriteLock();

    sumRdDelay = 0;
    sumWrDelay = 0;

    if ( queueLock.numReadLock() > 0 )
	sumRdDelay = queueLock.sumReadDelay() / LCKMNG_DELRES;
    if ( queueLock.numWriteLock() > 0 )
	sumWrDelay = queueLock.sumWriteDelay() / LCKMNG_DELRES;

}

==> cego-2.6.2.tar.gz: cego/src/CegoDbThreadPool.cc <==
///////////////////////////////////////////////////////////////////////////////
//                                                         
// CegoDbThreadPool.cc  
// -------------------                                                     
// Cego db threadpool class implementation
//                                                         
// Design and Implementation by Bjoern Lemke
//               
// (C)opyright 2000-2010 Bjoern Lemke
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: CegoDbThreadPool
// 
// Description: 
//
// Status: QG-2.5.5
//
///////////////////////////////////////////////////////////////////////////////

#include "Exception.h"
#include "Host.h"
#include "NetHandler.h"
#include "Net.h"
#include "Sleeper.h"
#include "ThreadLock.h"

#include "XMLSuite.h"

#include "CegoXMLdef.h"
#include "CegoDbThreadPool.h"

#include "CegoDefs.h"

static ThreadLock queueLock("DBQUEUE"); 

CegoDbThreadPool::CegoDbThreadPool() : Thread()
{
    queueLock.init(LCKMNG_LOCKWAITDELAY, LCKMNG_COUNTSTAT);

}

CegoDbThreadPool::CegoDbThreadPool(int poolLimit, CegoDatabaseManager *pDBMng)  : Thread()
{
    queueLock.init(LCKMNG_LOCKWAITDELAY, LCKMNG_COUNTSTAT);

    _poolLimit = poolLimit;

    pDBMng->getDataPort(_dataPortNo);

    // _dataPortNo = pDoc->getRootElement()->getAttributeValue(XML_DBPORT_ATTR).asInteger();

    _threadId = (long*)malloc(_poolLimit * sizeof(long));
    _numRequest = (long*)malloc(_poolLimit * sizeof(long));
    _threadState = (ThreadState*)malloc(_poolLimit * sizeof(ThreadState));
    _threadList = (CegoDbThread**)malloc(_poolLimit * sizeof(CegoDbThread*));

    _terminated=false;

    long i=0;
    while ( i < _poolLimit )
    {
	
	_threadState[i] = READY;
	_threadList[i]= new CegoDbThread(this, pDBMng);
	_numRequest[i]=0;
	_threadId[i]=i;
	_threadList[i]->start(&_threadId[i]);
	
	i++;
    }
    
    _pDBMng = pDBMng;

    _modId = _pDBMng->getModId("CegoDbThreadPool");
}


CegoDbThreadPool::~CegoDbThreadPool()  
{
    _terminated=true;

    long i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->abortSession();	
	i++;
    }

    _joined=false;
    int count=0;
    while ( _joined == false && count < POOL_TERMWAIT )
    {
	Sleeper s;
	s.secSleep(1);
	count++;
    }
    
    if ( _joined )
    {
	_pDBMng->log(_modId, Logger::NOTICE, Chain("All db threads terminated"));
	join(getTid());
    }
    else
    {
	_pDBMng->log(_modId, Logger::NOTICE, Chain("Canceling hanging db sessions ..."));
	cancel();
    }
    i=0;
    while ( i < _poolLimit )
    {
	delete _threadList[i];
	i++;
    }
    delete _numRequest;
    delete _threadId;
    delete _threadState;
	    
}

void  CegoDbThreadPool::loadObjects(int tabSetId)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->loadObjects(tabSetId);
	i++;
    }
}

void  CegoDbThreadPool::unloadObjects(int tabSetId)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->unloadObjects(tabSetId);
	i++;
    }
}

void CegoDbThreadPool::removeObject(int tabSetId, const Chain& objName, CegoObject::ObjectType type)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->removeObject(tabSetId, objName, type);
	i++;
    }    
}


void CegoDbThreadPool::loadObject(int tabSetId, const Chain& objName, CegoObject::ObjectType type)
{
    int i=0;
    while ( i < _poolLimit )
    {
	_threadList[i]->loadObject(tabSetId, objName, type);
	i++;
    }    
}

int CegoDbThreadPool::getPoolLimit() const
{
    return _poolLimit;
}

void CegoDbThreadPool::getThreadInfo(long i, long& numRequest, CegoDbThreadPool::ThreadState& state, Chain& action)
{
    state = _threadState[i];
    numRequest = _numRequest[i];    
    action = _threadList[i]->lastAction();
}

void CegoDbThreadPool::setState(long i, CegoDbThreadPool::ThreadState state)
{
    _threadState[i] = state;
}

void CegoDbThreadPool::incNumRequest(long i)
{
    _numRequest[i]++;
}

void CegoDbThreadPool::setTid(long i, long tid)
{
    _threadId[i] = tid;
}

void  CegoDbThreadPool::abortThread(long i)
{
    _threadList[i]->abortSession();
}

void* CegoDbThreadPool::job(void* arg)
{

    try {
    	
	
	Net net( NETMNG_MSG_BUFLEN, NETMNG_SIZEBUFLEN );
	
	// cout << "Serving portno " << _portNo << endl;
	net.serve(_dataPortNo);
	
	while ( ! _terminated ) 
	{
	    
	    // take the mutex to avoid cpu load during wait position
	
	    bool noReq=false;
	    lockQueue();
	    if ( _requestQueue.Size() == 0  )  noReq=true;
	    unlockQueue();
	    	    
	    if ( noReq )
		lockQueue();
	    
	    NetHandler* pHandle;

	    try {

		pHandle = net.nextRequest( NETMNG_SELECT_TIMEOUT );

	    }
	    catch ( Exception e )
	    {
		Chain msg;
		e.pop(msg);
		_pDBMng->log(_modId, Logger::LOGERR, Chain("Error on incoming connection : ") + msg);
	    }
	    
	    if ( noReq )
	    {
		unlockQueue();
		
		// struct timespec delay;	
		// delay.tv_sec = 0; 
		// delay.tv_nsec = NETMNG_QUEUE_DELAY;
		// nanosleep(&delay, NULL);

		Sleeper s;
		s.nanoSleep(NETMNG_QUEUE_DELAY);
	    }
	    
	    if ( pHandle )
	    {
		// pHandle->setSendTimeout(NETMNG_SEND_TIMEOUT);
		// pHandle->setRecvTimeout(NETMNG_RECV_TIMEOUT);

		lockQueue();
		_requestQueue.Insert(pHandle);
		unlockQueue();
	    }
	}
	long i=0;
	while ( i < _poolLimit )
	{
	    _pDBMng->log(_modId, Logger::DEBUG, Chain("Waiting for db thread with tid ") + Chain(_threadList[i]->getTid()) + Chain(" to terminate"));
	    join(_threadList[i]->getTid());
	    i++;
	}
	_joined=true;
    }
    catch ( Exception e)
    {
	Chain msg;
	e.pop(msg);	
	_pDBMng->log(_modId, Logger::LOGERR, Chain("Db Thread start failed : ") +  msg);	
	_terminated=true;
    }

    // waiting for db threads to terminate
    
    return 0;
}

bool CegoDbThreadPool::isTerminated()
{
    return _terminated;
}

NetHandler* CegoDbThreadPool::nextRequest()
{

    lockQueue();

    NetHandler **pRequest = _requestQueue.First();

    if ( pRequest )
    {

	NetHandler *pN = *pRequest;
	_requestQueue.Remove(*pRequest);
	unlockQueue();

	return pN; 
    }
    else
    {
	unlockQueue();
	return 0;
    }    
}

void CegoDbThreadPool::lockQueue()
{
    queueLock.writeLock();
}

void CegoDbThreadPool::unlockQueue()
{
    queueLock.unlock();
}


void CegoDbThreadPool::getLockStat(Chain& lockName, long& lockCount, long &numRdLock, long &numWrLock, long &sumRdDelay, long &sumWrDelay)
{
    lockName = queueLock.getId();
    lockCount = queueLock.numLockTry();

    numRdLock = queueLock.numReadLock();
    numWrLock = queueLock.numWriteLock();

    sumRdDelay = 0;
    sumWrDelay = 0;

    if ( queueLock.numReadLock() > 0 )
	sumRdDelay = queueLock.sumReadDelay() / LCKMNG_DELRES;
    if ( queueLock.numWriteLock() > 0 )
	sumWrDelay = queueLock.sumWriteDelay() / LCKMNG_DELRES;

}
